# -*- coding: utf-8 -*-

"""
Created on 2018-03-22

Script to download and process Esther Richards letters from Mt Holyoke Archive online

Script Author/Consultant:Nathaniel Porter (ndporter@vt.edu)
Paper Author: Katherine Cottle

Process:
    1. Dowload all letters in erichards archive related to E Richards (https://www.mtholyoke.edu/~dalbino/letters/women/erichards.html)
    2. Create dataframe with date/addressee/fileid of each letter from html
    3. Create and save matrix of word occurrence for each document in the corpus
    4. Use Latent Dirichlet analysis to analyze groups of words tending to occur frequently together and summarize

"""

#Import necessary packages
import pandas as pd #, numpy as np
from os import chdir,path
#from urllib.request import urlopen
from glob import glob
#import networkx as nx
from scipy import sparse #, io
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import LatentDirichletAllocation
#from wordcloud import WordCloud, STOPWORDS
#import matplotlib.pyplot as plt


############################
#Defined functions
############################
#Functions
#print top words and weights
def print_top_words(model, feature_names, n_top_words):
    for topic_id, topic in enumerate(model.components_):
        print('\nTopic %d:' % int(topic_id)) 
        print(''.join([feature_names[i] + ' ' + str(round(topic[i], 2))
              +' | ' for i in topic.argsort()[:-n_top_words - 1:-1]]))
#save top n words and weights to csv
"""
args:
    file: output csv file
    model: topic model
    feature_names: list of feature names
    n_top_words: number of words per topic to include (beginning from highest weight)
    topics: 
        'all' (default) includes all topics in model OR
        single topic id number (int) for file
    towordcloudsdotcom:
        True is formatted output to use with wordclouds.com (weight\tword\tcolor\turl)
            color and url are optional and site auto-detects and CURRENTLY NOT SUPPORTED IN DEFINITION
        False is csv format (topic,word,weight) to use as d3 etc input
        
usage:
    with topics == 'all': saves single csv of n top words per topic with weights
    with specified topic: saves one csv per topic (to use for WordCloud etc), recommended to use list comprehension to create multiple files
"""
def save_top_words(file, model, feature_names, n_top_words=20, topics='all', towordcloudsdotcom=False):
    with open(file, mode='w') as out:
        if towordcloudsdotcom:
            out.write('weight\tword\n')
            if topics=='all':
                print('Cannot combine all topics with towordcloudsdotcom')
                exit
            else:
                for topic_id, topic in enumerate(model.components_):
                    if str(topic_id) == topics:
                        out.write(''.join([str(int(round(topic[i],2)*100))+'\t'+feature_names[i]+'\n'
                                  for i in topic.argsort()[:-n_top_words - 1: -1]]))
                    else:
                        pass
               
        else:
            out.write('topic,word,weight\n')
            if topics=='all':
                for topic_id, topic in enumerate(model.components_):
                    out.write(''.join([str(topic_id)+','+feature_names[i]+','+str(round(topic[i],4))+'\n'
                              for i in topic.argsort()[:-n_top_words - 1: -1]]))
            else:
                for topic_id, topic in enumerate(model.components_):
                    if str(topic_id) == topics:
                        out.write(''.join([str(topic_id)+','+feature_names[i]+','+str(round(topic[i],4))+'\n'
                                  for i in topic.argsort()[:-n_top_words - 1: -1]]))
                    else:
                        pass
            

############################
#Executable script body
############################

###Preliminaries
#Create directory paths //EDIT THESE TO MATCH YOUR OWN LOCATION
workDir = r'C:\viral networks workshop\Cottle\LDA'
sourceDir = r'\sources'
outDir = r'\output'
sourcePath= workDir+sourceDir
outPath = workDir+outDir
chdir(workDir) #change path to scripts

"""
###Download all letter from archive based on url scheme
###Please obey all copyright, license, usage, and robots.txt restrictions when downloading sources
i = 0 #initialize counter
#download (by trying potential files)
while i < 100:
    i+=1
    fname = 'richards'+str(i)+'.html'
    url = r'http://www.mtholyoke.edu/~dalbino/letters/text/'+fname
    try: 
        html = urlopen(url)
        read = html.read()
        with open(sourcePath+'\\'+fname, 'w') as fid:
            fid.write(str(read))
    except Exception as e:
        print(str(e)+' for '+url)
"""
    
###Make list of files
filelist = glob(path.join(sourcePath,'*.html'))
#remove Cosmo article (#91, not correspondence) and letter to E Richards (#77, different recipient)
remNums = [91,77]
for num in remNums:
    f = sourcePath+r'\richards'+str(num)+'.html'
    while f in filelist: filelist.remove(f)

###Create text corpus
"""
docs: full text of correspondence documents
dates: list of sent dates
"""
##dates (ordered based on filelist ordered; link using position)
dates = []
for f in filelist:
    with open(f, 'r+') as letter:
        l = letter.read()
        head, sep, tail = l.partition('written on ')
        head, sep, tail = tail.partition('<')
        dates.append(head)
#note this lists leaves date as string not in datetime

###Create a matrix counting occurrence of each word in each document
docs = []
for f in filelist:
    with open(f, 'r+') as letter:
        l = letter.read()
        pars = l.replace(r'\r','').split(r'<P>') #parse each paragraph
        ltext = ''
        for par in pars:
            if r'<' not in par and len(par)>25:
                ltext = ltext+' '+par.replace('\\','')
    docs.append(ltext)
#Save total frequency of each word based on text
with open(outPath+r'\all_words_for_cloud.csv', 'w') as myfile:
    for item in docs:
        myfile.write(item)
        myfile.write('\n')
                
##Create a matrix counting occurrence of each word in each document
#Use options on CountVectorizer to remove stopwords and accents
count_model = CountVectorizer(strip_accents='unicode',stop_words='english',ngram_range=(1,1)) # default unigram model
X = count_model.fit_transform(docs) #creates sparse matrix with counts of each token within each row
names = count_model.get_feature_names() #ordered list to recover names from matrix

"""
##Create word cloud for everything
text = ' '.join(docs)
stopwords=set(STOPWORDS)
wc = WordCloud(background_color="white", max_words=200, stopwords=stopwords)
wc.generate(text)
# store to file
wc.to_file(path.join(outPath,"allRichardsCloud.png"))
# show
plt.imshow(wc, interpolation='bilinear')
plt.axis("off")
plt.savefig(outPath+r'\test.svg',format='svg') 
#note this SVG in name only; it just embeds the PNG in an SVG canvas
#SVG wordclouds used in figures for book created using https://wordclouds.com/

##Use matrix multiplication to convert a two-mode network (docs* words) into symmetrical W*W or D*D matrices
#multiplication has cols equal to first matrix, rows equal to second
Xc = (X.T * X) # this is co-occurrence matrix in sparse csr format 
Xc.setdiag(0) # this sets co-occurence of the same word to 0; otherwise it would be the sum of the squares of frequency in each

##Import data to networkx (Python network package)
G = nx.from_scipy_sparse_matrix(Xc) #create graph object
#add node labels
nodeLabelsMap = {} #gen empty dictionary
for n in range(len(names)):
    nodeLabelsMap[str(n)] = names[n] #add each mapping to dict
G = nx.relabel_nodes(G, nodeLabelsMap)

#Basic visualization
pos = nx.spring_layout(G)
for node,(x,y) in pos.items():
    G.node[node]['x'] = float(x)
    G.node[node]['y'] = float(y)
#nx.draw_networkx(G, pos)

##Export data to gephi as csv
pandasG = nx.to_pandas_adjacency(G)
pandasG.to_csv(outPath+r'/richardsTextNetwork.csv',sep='\t')
pd.DataFrame(names).to_csv(outPath+r'/richardsNodeNames.csv',sep='\t')

##Export as edgelist for Cytoscape (works but nodes don't seem to match well)
io.mmwrite(outPath+r'/richardsTextNetwork.mtx',Xc) #writes to matrix market format

"""

####Latent Dirichlet Allocation

#Initialize and Fit Model
lda = LatentDirichletAllocation(n_components=15,random_state=42,learning_method='batch') #refining n_components and min/max_df (below) are case by case
#lda.fit(X)
#Print top terms for each component
#print("\nTopics in LDA model:")
#print_top_words(lda, names, 20)
#Too many common terms in top lists - normalize using tf-idf

###Vectorize using tf-idf
tfidf_vectorizer = TfidfVectorizer(max_df=0.75, min_df=5,
                                   stop_words='english')
tfidf = tfidf_vectorizer.fit_transform(docs)
tfidf_feature_names = tfidf_vectorizer.get_feature_names()
lda.fit(tfidf)
print("\nTopics in TF-IDF LDA model:")
print_top_words(lda, tfidf_feature_names, 20)

###Save topic-word-weight tables
save_top_words(outPath+r'\all_words_all_topics.csv',lda,tfidf_feature_names,n_top_words=len(tfidf_feature_names)) #all words all topics
save_top_words(outPath+r'\top20_words_all_topics.csv',lda,tfidf_feature_names) #top 20 by topic

#per topic
def list_topics(model):
    temp = []
    for topic_id,topic in enumerate(model.components_):
        temp.append(str(topic_id))
    return temp
for topic in list_topics(lda):
    save_top_words(outPath+'\\topic'+topic+'.csv',lda,tfidf_feature_names,topics=topic)

###Generate word clouds for individual topics
#Use https://wordclouds.com/ to generate true SVG word clouds
for topic in list_topics(lda):
    save_top_words(outPath+'\\wordclouds\\topic'+topic+'.csv',lda,tfidf_feature_names, n_top_words=20, topics=topic, towordcloudsdotcom=True)

#Document-topic distribution
doc_topic_distrib = lda.transform(tfidf)
dtd = pd.DataFrame(doc_topic_distrib)
#Include only ties > 0.1
X= sparse.csr_matrix(doc_topic_distrib)
Xc = (X.T * X) # this is co-occurrence matrix in sparse csr format 
#leave diagonal values as self-ties (signifies how strongly linked topic is)
#print(Xc.todense()) # print out matrix in dense format (topic * topic)
temp=Xc.toarray() #convert scipy matrix back to np array
"""
#Saves topic-topic network to csv
args:
        file: location and name of output file
        model: defined model definition
        data: vector of word weights (must be same as used for model last fit)
        selftie: True includes self-ties to represent concentration of topic in specific docs; False exclude (zero diagonal)
"""
def save_ttt_to_csv(file,model,data,selftie=True):
    dtd=model.transform(data) #create doc-topic-distribution
    temp = sparse.csr_matrix(dtd) #transform to sparse matrix
    ttt = (temp.T*temp) #convert from dtd to ttt
    temp = ttt.toarray() #convert to array
    with open(file, 'w') as out:
        out.write('source,target,value\n')
        for i in range(len(temp)):
            for j in range(len(temp)):
                if (selftie==False and i == j) or i>j: #always removes second identical tie (symmetric); removes selfties on args
                    pass
                else:
                    out.write(str(i)+','+str(j)+','+str(round(temp[i,j],4))+'\n')
save_ttt_to_csv(outPath+'\\ttt.csv',lda,tfidf,selftie=True) #save with self-ties
save_ttt_to_csv(outPath+'\\ttt_no_self.csv',lda,tfidf,selftie=False) #save with zero-diagonal